"""File 18ordinalLogit.py

:author: Michel Bierlaire, EPFL
:date: Mon Sep  9 08:08:40 2019

 Example of an ordinal logit model.
 This is just to illustrate the syntax, as the data are not ordered. 
 But the example assume, for the sake of it, that they are 1->2->3
 Three alternatives: Train, Car and Swissmetro
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.expressions import Beta, DefineVariable, log, Elem

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)

# Parameters for the ordered logit. 
# tau1 <= 0
tau1	 = Beta('tau1',-1,None,0,0)
# delta2 >= 0
delta2	 = Beta('delta2',2,0,None,0)
tau2 = tau1 + delta2


# Definition of new variables
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)

#  Utility
U = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED 

# Associate each discrete indicator with an interval.
#   1: -infinity -> tau1
#   2: tau1 -> tau2
#   3: tau2 -> +infinity
ChoiceProba = {
    1: 1-dist.logisticcdf(U-tau1),
    2: dist.logisticcdf(U-tau1)- dist.logisticcdf(U-tau2),
    3: dist.logisticcdf(U-tau2) }

# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = log(Elem(ChoiceProba,CHOICE))

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
logger.setSilent()
#logger.setWarning()
#logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "18ordinalLogit"

# Estimate the parameters
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)






# Translated to .py by Meritxell Pacheco (December 2016)
# Adapted to PandasBiogeme by Nicola Ortelli (November 2019)

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.optimization as opt
from biogeme.expressions import Beta, DefineVariable, bioDraws, MonteCarlo



import biogeme.messaging as msg

logger = msg.bioMessage()
#logger.setDetailed()
logger.setDebug()


df = pd.read_csv("swissmetro.dat", sep = '\t')
database = db.Database("swissmetro", df)

globals().update(database.variables)


# Exclude data
exclude = (( PURPOSE != 1 ) * ( PURPOSE != 3 ) + ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR_mean = Beta('ASC_CAR_mean', 0, None, None, 0)
ASC_CAR_std  = Beta('ASC_CAR_std',  0, None, None, 0)
ASC_SM_mean  = Beta('ASC_SM_mean',  0, None, None, 0)
ASC_SM_std   = Beta('ASC_SM_std',   0, None, None, 0)
ASC_TRAIN    = Beta('ASC_SBB',      0, None, None, 1)

B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)
B_HE   = Beta('B_HE',   0, None, None, 0)

# Random parameters
ASC_CAR_random = ASC_CAR_mean + ASC_CAR_std * bioDraws('ASC_CAR_random', 'NORMAL')
ASC_SM_random  = ASC_SM_mean  + ASC_SM_std  * bioDraws('ASC_SM_random',  'NORMAL')

# Definition of new variables
TRAIN_COST = DefineVariable('TRAIN_COST', TRAIN_CO * ( GA == 0 ), database)
SM_COST    = DefineVariable('SM_COST', SM_CO * ( GA == 0 ), database)

# Utilities
V_TRAIN = ASC_TRAIN      + B_TIME * TRAIN_TT + B_COST * TRAIN_COST + B_HE * TRAIN_HE
V_SM    = ASC_SM_random  + B_TIME * SM_TT    + B_COST * SM_COST    + B_HE * SM_HE
V_CAR   = ASC_CAR_random + B_TIME * CAR_TT   + B_COST * CAR_CO

V = {1: V_TRAIN, 2: V_SM, 3: V_CAR}
av = {1: TRAIN_AV, 2: SM_AV, 3: CAR_AV}

# Choice model estimation
prob = models.logit(V, av, CHOICE)
logprob = log(MonteCarlo(prob))
biogeme = bio.BIOGEME(database, logprob, numberOfDraws = 100)
biogeme.modelName = "Mixture Heteroskedastic"
#results = biogeme.estimate(algorithm=opt.newtonTrustRegionForBiogeme)
results = biogeme.estimate()

# Results
pandasResults = results.getEstimatedParameters()
print(pandasResults)
print(f"Nbr of observations: {database.getNumberOfObservations()}")
print(f"LL(0) = {results.data.initLogLike:.3f}")
print(f"LL(beta) = {results.data.logLike:.3f}")
print(f"rho bar square = {results.data.rhoBarSquare:.3g}")
print(f"Output file: {results.data.htmlFileName}")



import unittest
import random as rnd
import biogeme.biogeme as bio
import biogeme.database as db
import pandas as pd
import numpy as np
from biogeme.expressions import *

class testBiogeme(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        rnd.seed(90267)
        df = pd.DataFrame({'Person':[1,1,1,2,2],
                   'Exclude':[0,0,1,0,1],
                   'Variable1':[1,2,3,4,5],
                   'Variable2':[10,20,30,40,50],
                   'Choice':[1,2,3,1,2],
                   'Av1':[0,1,1,1,1],
                   'Av2':[1,1,1,1,1],
                   'Av3':[0,1,1,1,1]})
        myData = db.Database('test',df)
        
        Variable1=Variable('Variable1')
        Variable2=Variable('Variable2')
        beta1 = Beta('beta1',-1.0,-3,3,0)
        beta2 = Beta('beta2',2.0,-3,10,0)
        likelihood = -beta1**2 * Variable1 - exp(beta2*beta1) * Variable2 - beta2**4
        simul = beta1 / Variable1 + beta2 / Variable2
        dictOfExpressions = {'loglike':likelihood,'beta1':beta1,'simul':simul}

        self.myBiogeme = bio.BIOGEME(myData,dictOfExpressions)
        self.myBiogeme.modelName = 'simpleExample'

    def test_calculateInitLikelihood(self):
        res = self.myBiogeme.calculateInitLikelihood()
        self.assertAlmostEqual(res,-115.30029248549191,5)

    def test_calculateLikelihood(self):
        x = self.myBiogeme.betaInitValues
        xplus = [v+1 for v in x]
        res = self.myBiogeme.calculateLikelihood(xplus,scaled=False)
        self.assertEqual(res,-555)

    def test_calculateLikelihoodAndDerivatives(self):
        x = self.myBiogeme.betaInitValues
        xplus = [v+1 for v in x]
        f,g,h,bhhh = self.myBiogeme.calculateLikelihoodAndDerivatives(xplus,scaled=False,hessian=True,bhhh=True)
        f_true = -555.0
        g_true = [-450., -540.]
        h_true = [[-1350.,  -150.], [ -150.,  -540.]]
        bhhh_true = [[49500., 48600.], [48600., 58320.]]
        self.assertEqual(f_true,f)
        self.assertListEqual(g_true,g.tolist())
        self.assertListEqual(h_true,h.tolist())
        self.assertListEqual(bhhh_true,bhhh.tolist())

    def test_likelihoodFiniteDifferenceHessian(self):
        x = self.myBiogeme.betaInitValues
        xplus = [v+1 for v in x]
        h = self.myBiogeme.likelihoodFiniteDifferenceHessian(xplus)
        h_true = [[-1380.00020229,  -150.        ],  [ -150.0000451 ,  -540.00005396]]
        for row,row_true in zip(h,h_true):
            for col,col_true in zip(row,row_true):
                self.assertAlmostEqual(col,col_true,5)

    def test_checkDerivatives(self):
        f,g,h,gdiff,hdiff = self.myBiogeme.checkDerivatives()
        gdiff_true=[-5.42793187e-06,  2.60800035e-05]
        hdiff_true=[[-8.04552171e-06,  7.36597983e-09],[-1.61387920e-07,  2.22928137e-05]]
        for col, col_true in zip(gdiff,gdiff_true):
            self.assertAlmostEqual(col,col_true,5)
        for row,row_true in zip(hdiff,hdiff_true):
            for col,col_true in zip(row,row_true):
                self.assertAlmostEqual(col,col_true,5)

    def test_estimate(self):
        results = self.myBiogeme.estimate(bootstrap=10)
        self.assertAlmostEqual(results.data.logLike,-67.0654904797005,5)

    def test_simulate(self):
        results = self.myBiogeme.estimate()
        s = self.myBiogeme.simulate(results.getBetaValues())
        self.assertAlmostEqual(s.loc[0,'loglike'],-6.09223388037049,5)

    def test_confidenceIntervals(self):
        results = self.myBiogeme.estimate(bootstrap=10)
        drawsFromBetas = results.getBetasForSensitivityAnalysis(self.myBiogeme.freeBetaNames)
        s = self.myBiogeme.simulate(results.getBetaValues())
        left, right = self.myBiogeme.confidenceIntervals(drawsFromBetas)
        self.assertLess(left.loc[0,'loglike'],s.loc[0,'loglike'])
        self.assertGreater(right.loc[0,'loglike'],s.loc[0,'loglike'])

if __name__ == '__main__':
    unittest.main()



""" Defines the various arithmetic expressions accepted by Biogeme.

:author: Michel Bierlaire

:date: Tue Mar 26 16:47:49 2019

"""
import numpy as np
import biogeme.exceptions as excep
import biogeme.cbiogeme as cb
import biogeme.messaging as msg
logger = msg.bioMessage()        

def isNumeric(obj):
    """ Identifies if an object is numeric, that is int, float or bool.
 
    :param obj: any object
    :type obb: object
    """
    return isinstance(obj,(int,float,bool))

class Expression:
    """This is the general arithmetic expression in biogeme. It serves as a base class for concrete expressions.
    """
    def __init__(self):
        """ Constructor
        """
        ## Logger
        self.logger = msg.bioMessage()        
        ## Parent expression
        self.parent = None
        ## List of children expressions
        self.children = list()
        ## Indices of the elementary expressions (dict)
        self.elementaryExpressionIndex = None
        ## dict of free parameters
        self.allFreeBetas = None
        ## list of names of free parameters
        self.freeBetaNames = None
        ## dict of fixed parameters
        self.allFixedBetas = None
        ## list of names of fixed parameters
        self.fixedBetaNames = None
        ## dict of random variables
        self.allRandomVariables = None
        ## list of random variables names
        self.randomVariableNames = None
        ## dict of draws
        self.allDraws = None
        ## list of draw types
        self.drawNames = None
        
    def __repr__(self):
        """
        built-in function used to compute the "official" string reputation of an object
        
        :return: description of the expression
        :rtype: string
        """
        return self.__str__()
    
    def __add__(self,other):
        """ 
        Operator overloading. Generate an expression for addition. 

        :param other: expression to be added
        :type other: biogeme.expressions.Expression
   
        :return: self + other 
        :rtype: biogeme.expressions.Expression
        """
        return Plus(self,other)
    
    def __radd__(self,other):
        """ 
        Operator overloading. Generate an expression for addition. 

        :param other: expression to be added
        :type other: biogeme.expressions.Expression
   
        :return: other + self
        :rtype: biogeme.expressions.Expression
        """
        return Plus(other,self)
    
    
    def __sub__(self,other):
        """ 
        Operator overloading. Generate an expression for substraction. 

        :param other: expression to substract
        :type other: biogeme.expressions.Expression
   
        :return: self - other
        :rtype: biogeme.expressions.Expression
        """
        return Minus(self,other)
    
    def __rsub__(self,other):
        """ 
        Operator overloading. Generate an expression for substraction. 

        :param other: expression to be substracted
        :type other: biogeme.expressions.Expression
   
        :return: other - self
        :rtype: biogeme.expressions.Expression
        """
        return Minus(other,self)
    
    def __mul__(self,other):
        """ 
        Operator overloading. Generate an expression for multiplication. 

        :param other: expression to be multiplied
        :type other: biogeme.expressions.Expression

        :return: self * other
        :rtype: biogeme.expressions.Expression
        """
        return Times(self,other)
    
    def __rmul__(self,other):
        """ 
        Operator overloading. Generate an expression for multiplication. 

        :param other: expression to be multiplied
        :type other: biogeme.expressions.Expression

        :return: other * self
        :rtype: biogeme.expressions.Expression
        """
        return Times(other,self)
    
    def __div__(self,other):
        """ 
        Operator overloading. Generate an expression for division.
        
        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: self / other
        :rtype: biogeme.expressions.Expression
        """
        return Divide(self,other)
    
    def __rdiv__(self,other):
        """ 
        Operator overloading. Generate an expression for division.
        
        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: other / self
        :rtype: biogeme.expressions.Expression
        """
        return Divide(other,self)
    
    def __truediv__(self,other):
        """ 
        Operator overloading. Generate an expression for division.

        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: self / other
        :rtype: biogeme.expressions.Expression
        
        """
        return Divide(self,other)
    
    def __rtruediv__(self,other):
        """ 
        Operator overloading. Generate an expression for division.

        :param other: expression for division
        :type other: biogeme.expressions.Expression

        :return: other / self
        :rtype: biogeme.expressions.Expression
        """
        return Divide(other,self) 

    def __neg__(self):
        """ 
        Operator overloading. Generate an expression for unary minus.

        :return: -self
        :rtype: biogeme.expressions.Expression
        """
        return UnaryMinus(self)

    def __pow__(self,other):
        """ 
        Operator overloading. Generate an expression for power.

        :param other: expression for power
        :type other: biogeme.expressions.Expression

        :return: self ^ other
        :rtype: biogeme.expressions.Expression
        """
        return Power(self,other)

    def __rpow__(self,other):
        """ 
        Operator overloading. Generate an expression for power.

        :param other: expression for power
        :type other: biogeme.expressions.Expression

        :return: other ^ self
        :rtype: biogeme.expressions.Expression
        """
        return Power(other,self)

    def __and__(self, other):
        """ 
        Operator overloading. Generate an expression for logical and.

        :param other: expression for logical and
        :type other: biogeme.expressions.Expression

        :return: self and other
        :rtype: biogeme.expressions.Expression
        """
        return And(self,other)

    def __or__(self, other):
        """ 
        Operator overloading. Generate an expression for logical or.

        :param other: expression for logical or
        :type other: biogeme.expressions.Expression

        :return: self or other
        :rtype: biogeme.expressions.Expression
        """
        return Or(self,other)

    def __eq__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison.

        :param other: expression for equality
        :type other: biogeme.expressions.Expression

        :return: self == other
        :rtype: biogeme.expressions.Expression
        """
        return Equal(self,other)

    def __ne__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison.

        :param other: expression for difference
        :type other: biogeme.expressions.Expression

        :return: self != other
        :rtype: biogeme.expressions.Expression
        """
        return NotEqual(self,other)
    
    def __le__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison.

        :param other: expression for less or equal
        :type other: biogeme.expressions.Expression

        :return: self <= other
        :rtype: biogeme.expressions.Expression
        """
        return LessOrEqual(self,other)

    def __ge__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison.

        :param other: expression for greater or equal
        :type other: biogeme.expressions.Expression

        :return: self >= other
        :rtype: biogeme.expressions.Expression
        """
        return GreaterOrEqual(self,other)
    
    def __lt__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison.

        :param other: expression for less than
        :type other: biogeme.expressions.Expression

        :return: self < other
        :rtype: biogeme.expressions.Expression
        """
        return Less(self,other)

    def __gt__(self, other):
        """ 
        Operator overloading. Generate an expression for comparison. 

        :param other: expression for greater than
        :type other: biogeme.expressions.Expression

        :return: self > other
        :rtype: biogeme.expressions.Expression
        """
        return Greater(self,other)

    def _prepareFormulaForEvaluation(self,database):
        """ Extract from the formula the elementary expressions (parameters,
        variables, random parameters) and decide a numbering convention.
        """ 

        self.variableNames = list(database.data.columns.values)

        self.elementaryExpressionIndex,self.allFreeBetas,self.freeBetaNames,self.allFixedBetas,self.fixedBetaNames,self.allRandomVariables,self.randomVariableNames,self.allDraws,self.drawNames = defineNumberingOfElementaryExpressions([self],self.variableNames)

        
        ## List of ids of the free beta parameters (those to be estimated)
        self.betaIds = [i for i in range(len(self.freeBetaNames))]

        ## List of values of the free beta parameters (those to be estimated)
        self.freeBetaValues = [self.allFreeBetas[x].initValue for x in self.freeBetaNames]
        ## List of values of the fixed beta parameters (those to be estimated)
        self.fixedBetaValues = [self.allFixedBetas[x].initValue for x in self.fixedBetaNames]

    def getValue_c(self,database,numberOfDraws=1000):
        """
        Evaluation of the expression

        In Biogeme the complexity of some expressions requires a
        specific implementation, in C++. This function invokes the
        C++ code to evaluate the value of the expression for a
        series of entries in a database. Note that this function
        will generate draws if needed.

        
        :param database: database
        :type database:  biogeme.database.Database
        :param numberOfDraws: number of draws if needed by Monte-Carlo integration.
        :type numberOfDraws: int

        :return: a list where each entry is the result of applying the expression on one entry of the dsatabase. 
        :rtype: numpy.array
        """
        self._prepareFormulaForEvaluation(database)

        if database.isPanel():
            ## Object containing the C++ implementation used by Biogeme.
            theC = cb.pyPanelBiogeme()
            theC.setDataMap(database.individualMap)
        else:
            theC = cb.pyBiogeme()
        theC.setData(database.data)
        if self.allDraws:
            database.generateDraws(self.allDraws,self.drawNames,numberOfDraws)
            theC.setDraws(database.theDraws)

        result = theC.simulateFormula(self.getSignature(),self.freeBetaValues,self.fixedBetaValues,database.data)
        return result
            
    def setOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression.

        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool
        :return: returns a set with the beta parameters appearing in the expression.
        :rtype: set(biogeme.expressions.Expression)
        """
        s = set()
        for e in self.children:
            s = s.union(e.setOfBetas(free,fixed))
        return s


    def setOfVariables(self):
        """
        Extract the set of variables used in the expression.

        :return: returns a set with the variables appearing in the expression.
        :rtype: set(biogeme.expressions.Expression)
        """
        return set(self.dictOfVariables().keys())


    def dictOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression.

        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool

        :return: a dict with the beta parameters appearing in the expression, the keys being the names of the parameters.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        s = {}
        for e in self.children:
            d = e.dictOfBetas(free,fixed)
            s = dict(s,**d)
        return s

    def dictOfVariables(self):
        """Recursively extract the variables appearing in the expression, and
        store them in a dictionary.

        :return: returns a dict with the variables appearing in the expression the keys being their names.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        s = {}
        for e in self.children:
            d = e.dictOfVariables()
            s = dict(s,**d)
        return s
    

    
    def dictOfRandomVariables(self):
        """Recursively extract the random variables appearing in the expression, and store them in a dictionary.

        :return: returns a dict with the random variables appearing in
                 the expression the keys being their names.

        :rtype: dict(string:biogeme.expressions.Expression)
        """
        s = {}
        for e in self.children:
            d = e.dictOfRandomVariables()
            s = dict(s,**d)
        return s
    
    def getElementaryExpression(self,name):
        """ Return: an elementary expression from its name if it appears in the
        expression.

        :param name: name of the elementary expression.
        :type name: string

        :return: the expression if it exists. None otherwise.
        :rtype: biogeme.expressions.Expression
        """
        for e in self.children:
            if e.getElementaryExpression(name) is not None:
                return e.getElementaryExpression(name)
        return None
            
    def setRow(self,row):
        """This function identifies the row of the database from which the values of the variables must be obtained.

        :param row: id of the row.
        :type row: int

        """
        # Row of the database where the values of the variables must
        self._row = row
        for e in self.children:
            e.setRow(row)

    def dictOfDraws(self):
        """Recursively extract the random variables 
        (draws for Monte-Carlo)
        appearing in the expression, and store them in a dictionary.

        :return: dict where the keys are the random variables and the elements the type of draws
        :rtype: dict(string:string)
        """
        draws = {}
        for e in self.children:
            d = e.dictOfDraws()
            if d:
                draws = dict(draws,**d)
        return draws
        
    def setUniqueId(self,idsOfElementaryExpressions):
        """Provides a unique id to the elementary expressions.

        :param idsOfElementaryExpressions: dictionary mapping the name of the elementary expression with their id.
        :type idsOfElementaryExpressions: dict(string:int)
           
        """
        for e in self.children:
            e.setUniqueId(idsOfElementaryExpressions)

    def setSpecificIndices(self,
                           indicesOfFreeBetas,
                           indicesOfFixedBetas,
                           indicesOfRandomVariables,
                           indicesOfDraws):
        """Provides an index to all elementary expressions, specific to their type

        :param indicesOfFreeBetas: dictionary mapping the name of the
                               free betas with their index
        :type indicesOfFreeBetas: dict(string:int)

        :param indicesOfFixedBetas: dictionary mapping the name of the 
                                fixed betas with their index
        :type indicesOfFixedBetas: dict(string:int)

        :param indicesOfRandomVariables: dictionary mapping the name of the 
                                random variables with their index
        :type indicesOfRandomVariables: dict(string:int)
        :param indicesOfDraws: dictionary mapping the name of the draws with
                            their index
        :type indicesOfDraws: dict(string:int)

        """
        for e in self.children:
            e.setSpecificIndices(indicesOfFreeBetas,
                                 indicesOfFixedBetas,
                                 indicesOfRandomVariables,
                                 indicesOfDraws)

    def setVariableIndices(self,indicesOfVariables):
        """
        Provide an index to all variables

        :param indicesOfVariables: dictionary mapping the name of the 
                                variables with their index
        :type indicesOfVariables: dict(string:int)

        """
        for e in self.children:
            e.setVariableIndices(indicesOfVariables)
            
            
    def getClassName(self):
        """
        Obtain the name of the top class of the expression structure

        :return: the name of the class
        :rtype: string
        """
        n = type(self).__name__
        return n


    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signatures of all the children expressions,
            2. the name of the expression between < >
            3. the id of the expression between { }
            4. the number of children between ( )
            5. the ids of each children, preceeded by a comma.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        """
        listOfSignatures = []
        for e in self.children:
            listOfSignatures += e.getSignature()
        mysignature = f'<{self.getClassName()}>'
        mysignature += f'{{{id(self)}}}'
        mysignature += f'({len(self.children)})'
        for e in self.children:
            mysignature += f',{id(e)}'
        listOfSignatures += [mysignature.encode()]
        return listOfSignatures

    def isContainedIn(self,t):
        """Check if the expression is contained in an expression of type t.
        Typically, this would be used to check that a bioDraws
        expression is contained in a MonteCarlo expression. If not, it
        cannot be evaluated.

        Return: 
           bool.

        See: Expression.embedExpression
        """
        if self.parent is None:
            return False
        if self.parent.getClassName() == t:
            return True
        return self.parent.isContainedIn(t)
    
    def embedExpression(self,t):
        """Check if the expression contains an expression of type t.
        Typically, this would be used to check that a MonteCarlo
        expression contains a bioDraws expression. 

        Return: 
           bool.

        See: Expression.isContainedIn
        """
        if self.getClassName() == t:
            return True
        for e in self.children:
            if e.embedExpression(t):
                return True
        return False

    def countPanelTrajectoryExpressions(self):
        """ Count the number of times the PanelLikelihoodTrajectory 
        is used in the formula. It should trigger an error if it
        is used more than once.
        """
        nbr = 0
        for e in self.children:
            nbr += e.countPanelTrajectoryExpressions()
        return nbr
    
    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)
        """
        listOfErrors = []
        listOfWarnings = []

        for e in self.children:
            err,war = e.audit(database)
            listOfErrors += err
            listOfWarnings += war
        return listOfErrors,listOfWarnings

    def changeInitValues(self,betas):
        """Modifies the initial values of the Beta parameters.

        The fact that the parameters are fixed or free is irrelevant here.

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """

        for e in self.children:
            e.changeInitValues(betas)
        

class BinaryOperator(Expression):
    """
    Base class for arithmetic expressions that are binary operators.
    This expression is the result of the combination of two expressions,
    typically addition, substraction, multiplication or division.
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        Expression.__init__(self)
        if isNumeric(left):
            self.left = Numeric(left)
        else:
            self.left = left
        if isNumeric(right):
            self.right = Numeric(right)
        else:
            self.right = right
        self.left.parent = self
        self.right.parent = self
        self.children.append(self.left)
        self.children.append(self.right)
        
       
class Plus(BinaryOperator):
    """
    Addition expression
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} + {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.left.getValue() + self.right.getValue()

class Minus(BinaryOperator):
    """
    Substraction expression
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)

    def __str__(self):
        return f"({self.left} - {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.left.getValue() - self.right.getValue()
    
class Times(BinaryOperator):
    """
    Multiplication expression
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression

        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} * {self.right})"

    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.left.getValue() * self.right.getValue()
    

            
class Divide(BinaryOperator):
    """
    Division expression
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)

    def __str__(self):
        return f"({self.left} / {self.right})"
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.left.getValue() / self.right.getValue()
    

class Power(BinaryOperator):
    """
    Power expression
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)

    def __str__(self):
        return f"({self.left} ** {self.right})"
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.left.getValue() ** self.right.getValue()

class bioMin(BinaryOperator):
    """
    Minimum of two expressions
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"bioMin({self.left},{self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        if (self.left.getValue() <= self.right.getValue()):
            return self.left.getValue()
        else:
            return self.right.getValue()

class bioMax(BinaryOperator):
    """
    Maximum of two expressions
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"bioMax({self.left},{self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        if (self.left.getValue() >= self.right.getValue()):
            return self.left.getValue()
        else:
            return self.right.getValue()
        
        
class And(BinaryOperator):
    """
    Logical and 
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression

        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} and {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        if (self.left.getValue() == 0.0):
            return 0.0
        if (self.right.getValue() == 0.0):
            return 0.0
        return 1.0
    

class Or(BinaryOperator):
    """
    Logical or 
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} or {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        if (self.left.getValue() != 0.0):
            return 1.0
        if (self.right.getValue() != 0.0):
            return 1.0
        return 0.0
    

class Equal(BinaryOperator):
    """
    Logical equal 
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} == {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() == self.right.getValue() else 0
        return r
    

class NotEqual(BinaryOperator):
    """
    Logical not equal  
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} != {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() != self.right.getValue() else 0
        return r
    
class LessOrEqual(BinaryOperator):
    """
    Logical less or equal
    """
    def __init__(self,left,right):
        """ Constructor
       
        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression

        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} <= {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() <= self.right.getValue() else 0
        return r
    

class GreaterOrEqual(BinaryOperator):
    """
    Logical greater or equal
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} >= {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() >= self.right.getValue() else 0
        return r
    

class Less(BinaryOperator):
    """
    Logical less 
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} < {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() < self.right.getValue() else 0
        return r
    

class Greater(BinaryOperator):
    """
    Logical greater 
    """
    def __init__(self,left,right):
        """ Constructor

        :param left: first arithmetic expression 
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression 
        :type right: biogeme.expressions.Expression
        """
        BinaryOperator.__init__(self,left,right)
        
    def __str__(self):
        return f"({self.left} > {self.right})"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        r = 1 if self.left.getValue() > self.right.getValue() else 0
        return r
    

class UnaryOperator(Expression):
    """
    Base class for arithmetic expressions that are unary operators.

    Such an expression is the result of the modification of another
    expressions, typically changing its sign.
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        Expression.__init__(self)
        if isNumeric(child):
            self.child = Numeric(child)
        else :
            self.child = child
        self.child.parent = self
        self.children.append(self.child)
            

    
class UnaryMinus(UnaryOperator):
    """
    Unary minus expression
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"(-{self.child})"
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return - self.child.getValue() 

class MonteCarlo(UnaryOperator):
    """
    Monte Carlo integration
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"MonteCarlo({self.child})"
        
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors,listOfWarnings = self.child.audit(database)
        if not self.child.embedExpression("bioDraws"):
            theError = f"The argument of MonteCarlo must contain a bioDraws: {self}"
            listOfErrors.append(theError)
        if self.child.embedExpression("MonteCarlo"):
            theError = f"It is not possible to include a MonteCarlo statement in another one: {self}"
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings
        
class bioNormalPdf(UnaryOperator):
    """
    Probability density function of a normal random variable
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"bioNormalPdf({self.child})"
        
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
        

class bioNormalCdf(UnaryOperator):
    """
    Cumulative Distribution Function of a normal random variable
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"bioNormalCdf({self.child})"
        
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
        

class PanelLikelihoodTrajectory(UnaryOperator):
    """
    Likelihood of a sequences of observations for the same individual
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"PanelLikelihoodTrajectory({self.child})"
        
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
    def countPanelTrajectoryExpressions(self):
        """ Count the number of times the PanelLikelihoodTrajectory 
        is used in the formula. 
        """
        return 1+self.child.countPanelTrajectoryExpressions()
    
    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors = []
        listOfWarnings = []
        if not database.isPanel():
            theError = f"Expression PanelLikelihoodTrajectory can only be used with panel data. Use the statement database.panel('IndividualId') to declare the panel strucutre of the data: {self}"
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings
    


class exp(UnaryOperator):
    """
    exponential expression
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"exp({self.child})"
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return np.exp(self.child.getValue())
    

class log(UnaryOperator):
    """
    logarithm expression
    """
    def __init__(self,child):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        
    def __str__(self):
        return f"log({self.child})"
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return np.log(self.child.getValue())

    
class Derive(UnaryOperator):
    """
    Derivative with respect to an elementary expression
    """
    def __init__(self,child,name):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        """        
        UnaryOperator.__init__(self,child)
        # Name of the elementary expression by which the derivative is taken
        self.elementaryName = name
        # Unique ID of the expression
        self.elementaryIndex = None

    def setUniqueId(self,idsOfElementaryExpressions):
        """
        Provides a unique id to the elementary expressions.

        :param idsOfElementaryExpressions: dictionary mapping the name of the elementary expression with their id.
        :type idsOfElementaryExpressions: dict(string:int)
           
        """
        if self.elementaryName in idsOfElementaryExpressions:
            self.elementaryIndex = idsOfElementaryExpressions[self.elementaryName]
        else:
            raise excep.biogemeError(f"No index is available for elementary expression {self.elementName}.")
        self.child.setUniqueId(idsOfElementaryExpressions)
        
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signatures of the child expression,
            2. the name of the expression between < >
            3. the id of the expression between { }
            4. the id of the child, preceeded by a comma.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        listOfSignatures = []
        listOfSignatures += self.child.getSignature()
        mysignature = f'<{self.getClassName()}>'
        mysignature += f'{{{id(self)}}}'
        mysignature += f',{id(self.child)}'
        mysignature += f',{self.elementaryIndex}'
        listOfSignatures += [mysignature.encode()]
        return listOfSignatures
            
    def __str__(self):
        return "Derive({self.child},'{self.elementName}')"
        
    def getValue(self):
        """ Not implemented in Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
    

class Integrate(UnaryOperator):
    """
    Numerical integration
    """
    def __init__(self,child,name):
        """ Constructor

        :param child: first arithmetic expression 
        :type child: biogeme.expressions.Expression
        :param name: name of the random variable for the integration.
        :type name: string
        """        
        UnaryOperator.__init__(self,child)
        self.randomVariableName = name
        self.randomVariableIndex = None

    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors,listOfWarnings = self.child.audit(database)
        if not self.child.embedExpression("RandomVariable"):
            theError = f"The argument of Integrate must contain a RandomVariable: {self}"
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings

    def setUniqueId(self,idsOfElementaryExpressions):
        """
        Provides a unique id to the elementary expressions. Overloads the generic function

        :param idsOfElementaryExpressions: dictionary mapping the name of the elementary expression with their id.
        :type idsOfElementaryExpressions: dict(string:int)
           
        """
        if self.randomVariableName in idsOfElementaryExpressions:
            self.randomVariableIndex = idsOfElementaryExpressions[self.randomVariableName]
        else:
            raise excep.biogemeError(f"No index is available for random variable {self.randomVariableName}.")
        self.child.setUniqueId(idsOfElementaryExpressions)

    def setSpecificIndices(self,
                           indicesOfFreeBetas,
                           indicesOfFixedBetas,
                           indicesOfRandomVariables,
                           indicesOfDraws):
        """
        Provide an index to all elementary expressions, specific to their type
        Overloads the generic function.

        :param indicesOfFreeBetas: dictionary mapping the name of the
                               free betas with their index
        :type indicesOfFreeBetas: dict(string:int)

        :param indicesOfFixedBetas: dictionary mapping the name of the 
                                fixed betas with their index
        :type indicesOfFixedBetas: dict(string:int)

        :param indicesOfRandomVariables: dictionary mapping the name of the 
                                random variables with their index
        :type indicesOfRandomVariables: dict(string:int)
        :param indicesOfDraws: dictionary mapping the name of the draws with
                            their index
        :type indicesOfDraws: dict(string:int)

        """
        if self.randomVariableName in indicesOfRandomVariables:
            self.randomVariableIndex = indicesOfRandomVariables[self.randomVariableName]
        else:
            raise excep.biogemeError(f"No index is available for random variable {self.randomVariableName}. Known random variables: {indicesOfRandomVariables.keys()}")
        self.child.setSpecificIndices(indicesOfFreeBetas,
                                      indicesOfFixedBetas,
                                      indicesOfRandomVariables,
                                      indicesOfDraws)

        
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signatures of the child expression,
            2. the name of the expression between < >
            3. the id of the expression between { }, preceeded by a comma
            4. the id of the children, preceeded by a comma
            5. the index of the randon variable, preceeded by a comma

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        listOfSignatures = []
        listOfSignatures += self.child.getSignature()
        mysignature = f'<{self.getClassName()}>'
        mysignature += f'{{{id(self)}}}'
        mysignature += f',{id(self.child)}'
        mysignature += f',{self.randomVariableIndex}'
        listOfSignatures += [mysignature.encode()]
        return listOfSignatures
            
    def __str__(self):
        return f"Integrate({self.child},'{self.randomVariableName}')"
        
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")
    
class Elementary(Expression):
    """Elementary expression. 

    It is typically defined by a name appearing in an expression. It
    can be a variable (from the database), or a parameter (fixed or to
    be estimated using maximum likelihood), a random variable for
    numrerical integration, or Monte-Carlo integration.

    """
    def __init__(self,name):
        """ Constructor

        :param name: name of the elementary experession.
        :type name: string

        """
        Expression.__init__(self)
        self.name = name

        # The id should be unique for all elementary expressions
        # appearing in a given set of formulas.
        self.uniqueId = None

    def __str__(self):
        return self.name

    def getElementaryExpression(self,name):
        """Return: an elementary expression from its name if it appears in the
        expression. None otherwise.
        """
        if self.name == name:
            return self
        else:
            return None
    
    def setUniqueId(self,idsOfElementaryExpressions):
        """
        Provides a unique id to the elementary expressions. Overloads the
        generic function

        :param idsOfElementaryExpressions: dictionary mapping the name of the elementary expression with their id.
        :type idsOfElementaryExpressions: dict(string:int)

        """
        if self.name in idsOfElementaryExpressions:
            self.uniqueId = idsOfElementaryExpressions[self.name]
        else:
            raise excep.biogemeError(f"No index is available for expression {self.name}. List of available indices: {[n for n,i in idsOfElementaryExpressions.items() ]}")




    
class bioDraws(Elementary):
    """
    Draws for Monte-Carlo integration
    """
    def __init__(self,name,type):
        """ Constructor

        :param name: name of the random variable with a series of draws.
        :type name: string
        :param type: type of draws.
        :type type: string
        """        
        Elementary.__init__(self,name)
        self.type = type
        self.drawId = None

    def __str__(self):
        return f"bioDraws('{self.name}','{self.type}')"
    
    def getValue(self):
        """ Not implementedin Python. The C++ implementation must be used insteald

        Raises: biogeme.exceptions.biogemeError
        """
        raise excep.biogemeError("Not implemented in Python. Use the C++ implementation.")

    def setSpecificIndices(self,
                           indicesOfFreeBetas,
                           indicesOfFixedBetas,
                           indicesOfRandomVariables,
                           indicesOfDraws):
        """
        Provide an index to all elementary expressions, specific to their type
        Overloads the generic function.

        :param indicesOfFreeBetas: dictionary mapping the name of the
                               free betas with their index
        :type indicesOfFreeBetas: dict(string:int)

        :param indicesOfFixedBetas: dictionary mapping the name of the 
                                fixed betas with their index
        :type indicesOfFixedBetas: dict(string:int)

        :param indicesOfRandomVariables: dictionary mapping the name of the 
                                random variables with their index
        :type indicesOfRandomVariables: dict(string:int)
        :param indicesOfDraws: dictionary mapping the name of the draws with
                            their index
        :type indicesOfDraws: dict(string:int)


        """
        if self.name in indicesOfDraws:
            self.drawId = indicesOfDraws[self.name]
        else:
            raise excep.biogemeError(f"No index is available for draw {self.type}. Known types of draws: {indicesOfDraws.keys()}")

    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the name of the expression between < >
            2. the id of the expression between { }, preceeded by a comma
            3. the name of the draws
            4. the unique ID (preceeded by a comma),
            5. the draw ID (preceeded by a comma).

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        :raise biogeme.exceptions.biogemeError: if no id has been defined for elementary expression
        :raise biogeme.exceptions.biogemeError: if no id has been defined for draw
        """
        if self.uniqueId is None:
            raise excep.biogemeError(f"No id has been defined for elementary expression {self.name}.")
        if self.drawId is None:
            raise excep.biogemeError(f"No id has been defined for draw {self.name}.")
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f'"{self.name}",{self.uniqueId},{self.drawId}'
        return [signature.encode()]

    def dictOfDraws(self):
        """Recursively extract the random variables 
        (draws for Monte-Carlo).  Overloads the generic function.
        appearing in the expression, and store them in a dictionary.

        :return: dict where the keys are the random variables and the elements the type of draws. Here, contains only one element.
        :rtype: dict(string:string)
        """
        return {self.name: self.type}

    def setDrawIndex(self,idsOfDraws):
        """
        Provide an index to a series of draw for a random variable. Overload the generic function.

        :param idsOfDraws: dictionary mapping the name of the draws with their id.
        :type idsOfDraws: dict(string:int)
           
        """
        if self.name in idsOfDraws:
            self.drawId = idsOfDraws[self.name]
        else:
            raise excep.biogemeError(f"No id is available for draw {self.name}. List of available indices: {[n for n,i in idsOfDraws.items()]}")
            

    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors = []
        listOfWarnings = []
        if not self.isContainedIn("MonteCarlo"):
            theError = f'bioDraws expression must be embedded into a MonteCarlo: {self}'
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings
            
    
class Numeric(Expression):
    """
    Numerical expression for a simple number
    """
    def __init__(self,value):
        """ Constructor

        :param value: numerical value
        :type value: float
        """
        Expression.__init__(self)
        self.value = value
        
    def __str__(self):
        return "`"+str(self.value)+"`"
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.value
    

    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the name of the expression between < >
            2. the id of the expression between { }
            3. the value, preceeded by a comma.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f',{self.value}'
        return [signature.encode()]
        
class Variable(Elementary):
    """Explanatory variable

    This represents the explanatory variables of the choice
    model. Typically, they come from the data set.
    """
    def __init__(self,name):
        """ Constructor

        :param name: name of the variable.
        :type name: string
        """
        Elementary.__init__(self,name)
        ## Index of the variable
        self.variableId = None
    
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self._row[self.name]

    def dictOfVariables(self):
        """Recursively extract the variables appearing in the expression, and
        store them in a dictionary.

        Overload the generic function.

        :return: returns a dict with the variables appearing in the expression the keys being their names. Here, it contains only one element.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        return {self.name: self}


    def audit(self,database):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors = []
        listOfWarnings = []
        if (database is None):
            raise excep.biogemeError("The database must be provided to audit the variable.")
        if not self.name in database.data.columns:
            theError = f"Variable {self.name} not found in the database."
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings

    def setVariableIndices(self,indicesOfVariables):
        """
        Provide an index to all variables

        :param indicesOfVariables: dictionary mapping the name of the 
                                variables with their index
        :type indicesOfVariables: dict(string:int)
           
        """
        if self.name in indicesOfVariables:
            self.variableId = indicesOfVariables[self.name]
    
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the name of the expression between < >
            2. the id of the expression between { }
            3. the name of the variable,
            4. the unique ID, preceeded by a comma.
            5. the variabvle ID, preceeded by a comma.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        :raise biogeme.exceptions.biogemeError: if no id has been defined for elementary expression
        :raise biogeme.exceptions.biogemeError: if no id has been defined for variable 
        """
        if self.uniqueId is None:
            raise excep.biogemeError(f"No id has been defined for elementary expression {self.name}.")
        if self.variableId is None:
            raise excep.biogemeError(f"No id has been defined for variable {self.name}.")
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f'"{self.name}",{self.uniqueId},{self.variableId}'
        return [signature.encode()]

class DefineVariable(Variable):
    """Expression that defines a new variable and add a column in the database.

       This expression allows the use to define a new variable that
       will be added to the database. It avoids that it is
       recalculated each time it is needed.
    """
    def __init__(self,name,expression,database):
        """ Constructor

        :param name: name of the variable.
        :type name: string
        :param expression: formula that defines the variable
        :param type:  biogeme.expressions.Expression
        :param database: object identifying the database.
        :type database: biogeme.database.Database

        """
        Variable.__init__(self,name)
        if isNumeric(expression):
            database.addColumn(Numeric(expression),name)
        else:
            database.addColumn(expression,name)

class RandomVariable(Elementary):
    """
    Random variable for numerical integration
    """
    def __init__(self,name):
        """ Constructor

        :param name: name of the random variable involved in the integration.
        :type name: string.
        """
        Elementary.__init__(self,name)
        # Index of the random variable
        self.rvId = None

    def audit(self,database=None):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors = []
        listOfWarnings = []
        if not self.isContainedIn("Integrate"):
            theError = f"RandomVariable expression must be embedded into a integrate: {self}"
            listOfErrors.append(theError)
        return listOfErrors,listOfWarnings
    
    def dictOfRandomVariables(self):
        """Recursively extract the random variables appearing in the expression, and store them in a dictionary.

        Overloads the generic function.

        :return: returns a dict with the random variables appearing in
                 the expression the keys being their names.

        :rtype: dict(string:biogeme.expressions.Expression)
        """
        return {self.name: self}


    def setSpecificIndices(self,
                           indicesOfFreeBetas,
                           indicesOfFixedBetas,
                           indicesOfRandomVariables,
                           indicesOfDraws):
        """
        Provide an index to all elementary expressions, specific to their type
        Overloads the generic function.

        :param indicesOfFreeBetas: dictionary mapping the name of the
                               free betas with their index
        :type indicesOfFreeBetas: dict(string:int)

        :param indicesOfFixedBetas: dictionary mapping the name of the 
                                fixed betas with their index
        :type indicesOfFixedBetas: dict(string:int)

        :param indicesOfRandomVariables: dictionary mapping the name of the 
                                random variables with their index
        :type indicesOfRandomVariables: dict(string:int)
        :param indicesOfDraws: dictionary mapping the name of the draws with
                            their index
        :type indicesOfDraws: dict(string:int)


        """
        if self.name in indicesOfRandomVariables:
            self.rvId = indicesOfRandomVariables[self.name]
        else:
            raise excep.biogemeError(f"No index is available for random variable {self.name}. Known random variables: {indicesOfRandomVariables.keys()}")
    
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the name of the expression between < >
            2. the id of the expression between { }
            3. the name of the random variable,
            4. the unique ID, preceeded by a comma,
            5. the ID of the random variable.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        :raise biogeme.exceptions.biogemeError: if no id has been defined for elementary expression
        :raise biogeme.exceptions.biogemeError: if no id has been defined for random variable
        """
        if self.uniqueId is None:
            raise excep.biogemeError(f"No id has been defined for elementary expression {self.name}.")
        if self.rvId is None:
            raise excep.biogemeError(f"No id has been defined for random variable {self.name}.")
        
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f'"{self.name}",{self.uniqueId},{self.rvId}'
        return [signature.encode()]
    
class Beta(Elementary):
    """
    Unknown parameters to be estimated from data.
    """
    def __init__(self,name,value,lowerbound,upperbound,status):
        """ Constructor

        :param name: name of the parameter.
        :type name: string
        :param value: default value.
        :type value: float
        :param lowerbound: if different from None, imposes a lower bound on the value of the parameter during the optimization. 
        :type lowerbound: float
        :param upperbound: if different from None, imposes an upper bound on the value of the parameter during the optimization. 
        :type upperbound: float
        :param status: if different from 0, the parameter is fixed to its default value, and not modified by the optimization algorithm.
        :type status: int
        """

        if not isinstance(value, (int, float)):
            raise excep.biogemeError(f'The second parameter for {name} must be a float and not a {type(value)}: {value}')
        if not isinstance(name,str):
            raise excep.biogemeError(f'The first parameter must be a string and not a {type(name)}: {name}')
        Elementary.__init__(self,name)
        self.initValue = value
        self.lb = lowerbound
        self.ub = upperbound
        self.status = status
        self.betaId = None

    def __str__(self):
        return "{}({})".format(self.name,self.initValue)
        
    def setOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression. Overload the generic function. 
        
        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool
        :return: returns a set with the beta parameters appearing in the expression.
        :rtype: set(biogeme.expressions.Expression)

        """
        if fixed and self.status != 0:
            return set([self.name])
        elif free and self.status == 0:
            return set([self.name])
        else:
            return set()


    def setSpecificIndices(self,
                           indicesOfFreeBetas,
                           indicesOfFixedBetas,
                           indicesOfRandomVariables,
                           indicesOfDraws):
        """
        Provide an index to all elementary expressions, specific to their type

        :param indicesOfFreeBetas: dictionary mapping the name of the
                               free betas with their index
        :type indicesOfFreeBetas: dict(string:int)

        :param indicesOfFixedBetas: dictionary mapping the name of the 
                                fixed betas with their index
        :type indicesOfFixedBetas: dict(string:int)

        :param indicesOfRandomVariables: dictionary mapping the name of the 
                                random variables with their index
        :type indicesOfRandomVariables: dict(string:int)
        :param indicesOfDraws: dictionary mapping the name of the draws with
                            their index
        :type indicesOfDraws: dict(string:int)
           
        """

        if self.status != 0:
            if self.name in indicesOfFixedBetas:
                self.betaId = indicesOfFixedBetas[self.name]
            else:
                raise excep.biogemeError(f"No index is available for fixed parameter {self.name}. Known fixed parameters: {indicesOfFixedBetas.keys()}")
        else:
            if self.name in indicesOfFreeBetas:
                self.betaId = indicesOfFreeBetas[self.name]
            else:
                raise excep.biogemeError(f"No index is available for free parameter {self.name}. Known free parameters: {indicesOfFreeBetas.keys()}")
            
    def dictOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression. Overload the generic function. 

        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool

        :return: a dict with the beta parameters appearing in the expression, the keys being the names of the parameters.
        :rtype: dict(string:biogeme.expressions.Expression)
        """
        if fixed and self.status != 0:
            return {self.name:self}
        elif free and self.status == 0:
            return {self.name:self}
        else:
            return dict()
        
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        return self.initValue

    def changeInitValues(self,betas):
        """Modifies the initial values of the Beta parameters.

        The fact that the parameters are fixed or free is irrelevant here.

        :param betas: dictionary where the keys are the names of the
                      parameters, and the values are the new value for
                      the parameters.
        :type betas: dict(string:float)
        """
        
        if self.name in betas:
            self.initValue = betas[self.name]

    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the name of the expression between < >
            2. the id of the expression between { }
            3. the name of the parameter,
            4. the status between [ ]
            5. the unique ID,  preceeded by a comma
            6. the beta ID,  preceeded by a comma
        

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)

        :raise biogeme.exceptions.biogemeError: if no id has been defined for elementary expression
        :raise biogeme.exceptions.biogemeError: if no id has been defined for parameter
        """
        if self.uniqueId is None:
            raise excep.biogemeError(f"No id has been defined for elementary expression {self.name}.")
        if self.betaId is None:
            raise excep.biogemeError(f"No id has been defined for parameter {self.name}.")
        
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f'"{self.name}"[{self.status}],{self.uniqueId},{self.betaId}'
        return [signature.encode()]
    
class LogLogit(Expression) :
    """Expression capturing the logit formula. 

    It contains one formula for the target alternative, a dict of
    formula for the availabilities and a dict of formulas for the
    utilities
    """
    def __init__(self, util, av, choice) :
        """Constructor

        :param util: dictionary where the keys are the identifiers of
                     the alternatives, and the elements are objects
                     defining the utility functions.

        :type util: dict(int:biogeme.expressions.Expression)
    
        :param av: dictionary where the keys are the identifiers of
                   the alternatives, and the elements are object of
                   type biogeme.expressions.Expression defining the
                   availability conditions. If av is None, all the
                   alternatives are assumed to be always available

        :type av: dict(int:biogeme.expressions.Expression)

        :param choice: formula to obtain the alternative for which the
                       logit probability must be calculated.
        :type choice: biogeme.expressions.Expression

        """
        Expression.__init__(self)
        self.util = {}
        for i,e in util.items():
            if isNumeric(e):
                self.util[i] = Numeric(e)
            else:
                self.util[i] = e
        self.av = {}
        if av is None:
            self.av = {k:Numeric(1) for k,v in util.items()}
        else:
            for i,e in av.items():
                if isNumeric(e):
                    self.av[i] = Numeric(e)
                else:
                    self.av[i] = e
        if isNumeric(choice):
            self.choice = Numeric(choice)
        else:
            self.choice = choice

        self.choice.parent = self
        self.children.append(self.choice)
        for i,e in self.util.items():
            e.parent = self
            self.children.append(e)
        for i,e in self.av.items():
            e.parent = self
            self.children.append(e)

    def audit(self,database):
        """ Performs various checks on the expressions.

        :param database: database object  
        :type database: biogeme.database.Database
        
        :return: tuple listOfErrors,listOfWarnings
        :rtype: list(string),list(string)

        """
        listOfErrors = []
        listOfWarnings = []
        for e in self.children:
            err,war = e.audit(database)
            listOfErrors += err
            listOfWarnings += war

        
        if self.util.keys() != self.av.keys():
            theError = "Incompatible list of alternatives in logit expression. "
            consistent = False
            set = self.util.keys() - self.av.keys()
            if set:
                theError += "Id(s) used for utilities and not for availabilities: "+",".join(f'{str(k)} ' for k in set)
            set = self.av.keys() - self.util.keys()
            if set:
                theError += " Id(s) used for availabilities and not for utilities: "+",".join(f'{str(k)} ' for k in set )
            listOfErrors.append(theError)
        else:
            consistent = True
        listOfAlternatives = list(self.util)
        choices = database.valuesFromDatabase(self.choice)
        correctChoices = choices.isin(listOfAlternatives)
        indexOfIncorrectChoices = correctChoices.index[correctChoices == False].tolist()
        if indexOfIncorrectChoices:
            incorrectChoices = choices[indexOfIncorrectChoices]
            theError = "The choice variable [{}] does not correspond to a valid alternative for the following observations (rownumber[choice]): ".format(self.choice)+'-'.join('{}[{}]'.format(*t) for t in zip(indexOfIncorrectChoices,incorrectChoices))
            listOfErrors.append(theError)

        if consistent:
            choiceAvailability = database.checkAvailabilityOfChosenAlt(self.av,self.choice)
            indexOfUnavailableChoices = choiceAvailability.index[choiceAvailability == False].tolist()
            if indexOfUnavailableChoices:
                incorrectChoices = choices[indexOfUnavailableChoices]
                theError = "The chosen alternative is not available for the following observations (rownumber[choice]): ".format(self.choice)+'-'.join('{}[{}]'.format(*t) for t in zip(indexOfUnavailableChoices,incorrectChoices))
                listOfWarnings.append(theError)

                
        return listOfErrors,listOfWarnings
            
    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        choice = int(self.choice.getValue())
        if choice not in self.util:
            self.logger.warning("Choice is {}. List of alternatives is {}".format(choice,self.util.keys()))
            return np.nan
        if self.av[choice].getValue() == 0.0:
            return -np.log(0)
        Vchosen = self.util[choice].getValue()
        denom = 0.0
        for i,V in self.util.items():
            if self.av[i].getValue() != 0.0:
                denom += np.exp(V.getValue()-Vchosen)
        return -np.log(denom)
        



    def __str__(self):
        s = self.getClassName()
        first = True
        for i,e in self.util.items():
            if first:
                s += "{}:{}".format(int(i),e)
                first = False
            else :
                s += ",{}:{}".format(int(i),e)
        s += ")"
        return s
    
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signatures of all the children expressions,
            2. the name of the expression between < >
            3. the id of the expression between { }
            4. the number of alternatives between ( )
            5. the id of the expression for the chosen alternative, preceeded by a comma.
            6. for each alternative, separated by commas: 

                 a. the number of the alternative, as defined by the user,
                 b. the id of the expression for the utility,
                 c. the id of the expression for the availability condition.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        listOfSignatures = []
        for e in self.children:
            listOfSignatures += e.getSignature()
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += f'({len(self.util)})'
        signature += f',{id(self.choice)}'
        for i,e in self.util.items():
            signature += f',{i},{id(e)},{id(self.av[i])}'
        listOfSignatures += [signature.encode()]
        return listOfSignatures
        

class _bioLogLogit(LogLogit):
    """log of logit formula

    This expression captures the logarithm of the logit formula. It
    contains one formula for the target alternative, a dict of formula
    for the availabilities and a dict of formulas for the utilities It
    uses only the C++ implementation.
    """
    pass

class _bioLogLogitFullChoiceSet(LogLogit):
    """This expression captures the logarithm of the logit formula, where
 all alternatives are supposed to be always available. 

    It contains one formula for the target alternative and a dict of
    formulas for the utilities. It uses only the C++ implementation.

    """
    pass



class bioMultSum(Expression):
    """This expression returns the sum of several other expressions. 

    It is a generalization of 'Plus' for more than two terms
    """
    def __init__(self,listOfExpressions):
        """Constructor

        :param listOfExpressions: list of objects representing the
                                     terms of the sum.

        :type listOfExpressions: list(biogeme.expressions.Expression)

        """        
        Expression.__init__(self)
        if type(listOfExpressions) is dict:
            for k,e in listOfExpressions.items():
                if isNumeric(e):
                    theExpression = Numeric(e)
                    theExpression.parent = self
                    self.children.append(theExpression)
                else:
                    e.parent = self
                    self.children.append(e)
        elif type(listOfExpressions) is list:
            for e in listOfExpressions:
                if isNumeric(e):
                    theExpression = Numeric(e)
                    theExpression.parent = self
                    self.children.append(theExpression)
                else:
                    e.parent = self
                    self.children.append(e)
        else:
            raise excep.biogemeError("Argument of bioMultSum must be a dict or a list.")


    def getValue(self):

        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        result = 0.0
        for e in self.children:
            result += e.getValue()
        return result

    def __str__(self):
        first = True
        s = 'bioMultSum'
        for e in self.children:
            if first:
                s += "({}".format(e)
                first = False
            else:
                s += " + {}".format(e)
        s += ')'
        return s
    
class Elem(Expression):
    """This returns the element of a dictionary. The key is evaluated
    from an expression.
    """
    def __init__(self,dictOfExpressions,keyExpression):
        """Constructor
    
        :param dictOfExpressions: dict of objects with numerical keys. 
        :type dictOfExpressions: dict(int: biogeme.expressions.Expression)

        :param keyExpression: object providing the key of the element
                              to be evaluated.
        :type keyExpression: biogeme.expressions.Expression

        """
        Expression.__init__(self)
        
        self.dictOfExpressions = {}
        for k,v in dictOfExpressions.items():
            if isNumeric(v):
                self.dictOfExpressions[k] = Numeric(v)
            else:
                self.dictOfExpressions[k] = v
            self.dictOfExpressions[k].parent = self                
            self.children.append(self.dictOfExpressions[k])
                
        if type(keyExpression) is bool:
            self.keyExpression = Numeric(1) if keyExpression else Numeric(0)
        elif isNumeric(keyExpression):
            self.keyExpression = Numeric(keyExpression)
        else:
            self.keyExpression = keyExpression
        self.keyExpression.parent = self
        self.children.append(self.keyExpression)

    def getValue(self):
        """ Evaluates the value of the expression

        :return: value of the expression 
        :rtype: float
        """
        key = int(self.keyExpression.getValue())
        if key in self.dictOfExpressions:
            return self.dictOfExpressions[key].getValue()
        else:
            return 0.0

    def __str__(self):
        s = '{{'
        first = True
        for k,v in self.dictOfExpressions.items():
            if first:
                s += '{}:{}'.format(k,v)
                first = False
            else:
                s += ',{}:{}'.format(k,v)
        s += '}}[{}]'.format(self.keyExpression)
        return s

    
    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signature of the expression defining the key
            2. the signatures of all the children expressions,
            3. the name of the expression between < >
            4. the id of the expression between { }
            5. the number of elements between ( )
            6. the id of the expression defining the key
            7. for each element: the value of the key and the id of the expression, separated by commas.

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        listOfSignatures = []
        listOfSignatures += self.keyExpression.getSignature()
        for i,e in self.dictOfExpressions.items():
            listOfSignatures += e.getSignature()
        signature = '<{}>'.format(self.getClassName())
        signature += '{{{}}}'.format(id(self))
        signature += '({})'.format(len(self.dictOfExpressions))
        signature += ',{}'.format(id(self.keyExpression))
        for i,e in self.dictOfExpressions.items():
            signature += ',{},{}'.format(i,id(e))
        listOfSignatures += [signature.encode()]
        return listOfSignatures

class bioLinearUtility(Expression):
    """When the utility function is linear, it is expressed as a list of
    terms, where a parameter multiplies a variable.
    """
    def __init__(self,listOfTerms):
        """Constructor

        :param listOfTerms: a list of tuple. Each tuple contains first a beta parameter, second the name of a variable. 
        :type listOfTerms: list(biogeme.expressions.Expression,biogeme.expressions.Expression)

        :raises biogeme.exceptions.biogemeError: if the object is not
                        a list of tuples (parameter,variable)

        """
        Expression.__init__(self)

        theError = ""
        first = True
        
        for b,v in listOfTerms:
            if not isinstance(b,Beta):
                if first:
                    theError += "Each element of the bioLinearUtility must be a tuple (parameter,variable). "
                    first = False
                theError += f" Expression {b} is not a parameter."
            if not isinstance(v,Variable):
                if first:
                    theError += "Each element of the list should be a tuple (parameter,variable)."
                    first = False
                theError += f" Expression {v} is not a variable."
        if not first:
            raise excep.biogemeError(theError)
        
        self.betas, self.variables = zip(*listOfTerms)
        self.betas = list(self.betas)
        self.variables = list(self.variables)
        self.listOfTerms = list(zip(self.betas,self.variables))
        self.children += self.betas + self.variables

        
    def setOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression.

        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool
        :return: returns a set with the beta parameters appearing in the expression.
        :rtype: set(biogeme.expressions.Expression)

        """
        if free:
            return set(self.betas)
        else:
            return set()

    def dictOfBetas(self,free=True,fixed=False):
        """
        Extract the set of parameters from the expression.

        :param free: if True, the free parameters are included. Default: True.
        :type free: bool
        :param fixed: if True, the fixed parameters are included. Default: False.
        :type fixed: bool

        :return: a dict with the beta parameters appearing in the expression, the keys being the names of the parameters.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        names = dict() 
        freenames = {x.name:x for x in self.betas if x.status == 0}
        fixednames = {x.name:x for x in self.betas if x.status != 0}
        if free and fixed:
            allnames = {**freenames,**fixednames}
            return allnames
        if free:
            return freenames
        if fixed:
            return fixednames
        return dict()
    
    def dictOfVariables(self):
        """Recursively extract the variables appearing in the expression, and
        store them in a dictionary.

        Overloads the generic function.

        :return: returns a dict with the variables appearing in the expression the keys being their names.
        :rtype: dict(string:biogeme.expressions.Expression)

        """
        return {x.name: x for x in self.variables}

    def dictOfRandomVariables(self):
        """Recursively extract the random variables appearing in the expression, and store them in a dictionary.

        :return: returns a dict with the random variables appearing in
                 the expression the keys being their names.

        :rtype: dict(string:biogeme.expressions.Expression)
        """
        return dict()

    def dictOfDraws(self):
        """Recursively extract the random variables 
        (draws for Monte-Carlo).  Overloads the generic function.
        appearing in the expression, and store them in a dictionary.

        :return: dict where the keys are the random variables and the elements the type of draws. Here, returns an empty dict.
        :rtype: dict(string:string)
        """
        return dict()

    def getSignature(self):
        """The signature of a string characterizing an expression. 

        This is designed to be communicated to C++, so that the
        expression can be reconstructed in this environment.

        The list contains the following elements: 

            1. the signatures of all the children expressions,
            2. the name of the expression between < >
            3. the id of the expression between { }
            4. the number of terms in the utility ( )
            5. for each term: 
        
                a. the id of the beta parameter
                b. the unique id of the beta parameter
                c. the name of the parameter
                d. the id of the variable
                e. the unique id of the variable
                f. the name of the variable

        Consider the following expression: 

        .. math:: 2 \\beta_1  V_1 - \\frac{\\exp(-\\beta_2 V_2) }{ \\beta_3  (\\beta_2 \\geq \\beta_1)}.

        It is defined as::
        
            2 * beta1 * Variable1 - expressions.exp(-beta2*Variable2) / (beta3 * (beta2 >= beta1))  

        And its signature is::

            [b'<Numeric>{4780527008},2',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<Times>{4780526952}(2),4780527008,4780277152',
             b'<Variable>{4511837152}"Variable1",5,2',
             b'<Times>{4780527064}(2),4780526952,4511837152',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<UnaryMinus>{4780527120}(1),4780277656',
             b'<Variable>{4511837712}"Variable2",6,3',
             b'<Times>{4780527176}(2),4780527120,4511837712',
             b'<exp>{4780527232}(1),4780527176',
             b'<Beta>{4780277264}"beta3"[1],2,0',
             b'<Beta>{4780277656}"beta2"[0],1,1',
             b'<Beta>{4780277152}"beta1"[0],0,0',
             b'<GreaterOrEqual>{4780527288}(2),4780277656,4780277152',
             b'<Times>{4780527344}(2),4780277264,4780527288',
             b'<Divide>{4780527400}(2),4780527232,4780527344',
             b'<Minus>{4780527456}(2),4780527064,4780527400']

        :return: list of the signatures of an expression and its children.
        :rtype: list(string)
        """
        listOfSignatures = []
        for e in self.children:
            listOfSignatures += e.getSignature()
        signature = f'<{self.getClassName()}>'
        signature += f'{{{id(self)}}}'
        signature += '({})'.format(len(self.listOfTerms))
        for b,v in self.listOfTerms:
            signature += f',{id(b)},{b.uniqueId},{b.name},{id(v)},{v.uniqueId},{v.name}'
        listOfSignatures += [signature.encode()]
        return listOfSignatures
        
    
def defineNumberingOfElementaryExpressions(collectionOfFormulas,variableNames):
    """Provides indices for elementary expressions

    The numbering is done in the following order:
    
        (i) free betas, 
        (ii) fixed betas, 
        (iii) random variables for numrerical integration, 
        (iv) random variables for Monte-Carlo integration, 
        (v) variables

      The numbering convention will be performed for all expressions
      together, so that the same elementary expressions in several
      expressions will have the same index.

    :param collectionOfFormula: collection of Biogeme expressions.
    :type collectionOfFormula: list(biogeme.expressions.Expression)
    :param variableNames: list of the names of the variables
    :type variableNames: list(string)
    :return: dict,free,freeNames,fixed,fixedNames,rv,rvNames,draws,drawsNames where 

         - dict is a dictionary mapping the names of the elementary expressions with their index, 
         - free is a dict with the free betas, 
         - freeNames is a list of the names of the free betas,
         - fixed is a dict with the fixed betas, 
         - fixedNames is the list of the names of the fixed betas, 
         - rv is a dict with the random variables for numerical integration, 
         - rvNames is a list with their names,
         - draws is a dict of the draws, and 
         - drawsNames is a list with their names.
    """
    # Free parameters (to be estimated), sorted by alphatical order.
    allFreeBetas = dict()
    freeBetaIndex = {}
    for f in collectionOfFormulas:
        d = f.dictOfBetas(free=True,fixed=False)
        allFreeBetas = dict(allFreeBetas,**d)
    freeBetaNames = sorted(allFreeBetas)
    for i in range(len(freeBetaNames)):
        freeBetaIndex[freeBetaNames[i]] = i

    # Fixed parameters (not to be estimated), sorted by alphatical order.
    allFixedBetas = dict()
    fixedBetaIndex = {}
    for f in collectionOfFormulas:
        d = f.dictOfBetas(free=False,fixed=True)
        allFixedBetas = dict(allFixedBetas,**d)
    fixedBetaNames = sorted(allFixedBetas)
    for i in range(len(fixedBetaNames)):
        fixedBetaIndex[fixedBetaNames[i]] = i

    # Random variables for numerical integration
    allRandomVariables = dict()
    randomVariableIndex = {}
    for f in collectionOfFormulas:
        d = f.dictOfRandomVariables()
        allRandomVariables = dict(allRandomVariables,**d)
    randomVariableNames = sorted(allRandomVariables)
    for i in range(len(randomVariableNames)):
        randomVariableIndex[randomVariableNames[i]] = i

    # Draws
    allDraws = dict()
    drawIndex = {}
    for f in collectionOfFormulas:
        d = f.dictOfDraws()
        allDraws = dict(allDraws,**d)
    drawNames = sorted(allDraws)
    for i in range(len(drawNames)):
        drawIndex[drawNames[i]] = i

    
    # Variables
    variableIndex = {}
    for i in range(len(variableNames)):
        variableIndex[variableNames[i]] = i

    # Merge all the names
    allElementaryExpressions = freeBetaNames + fixedBetaNames + randomVariableNames + drawNames + variableNames
    if len(allElementaryExpressions) != len(set(allElementaryExpressions)):
        duplicates = set([x for x in allElementaryExpressions if allElementaryExpressions.count(x) > 1])
        raise excep.biogemeError(f"The following elementary expressions are defined more than once: {duplicates}.")

    elementaryExpressionIndex = {}
    for i in range(len(allElementaryExpressions)):
        elementaryExpressionIndex[allElementaryExpressions[i]] = i
    
    for f in collectionOfFormulas:
        f.setUniqueId(elementaryExpressionIndex)
        f.setSpecificIndices(freeBetaIndex,fixedBetaIndex,randomVariableIndex,drawIndex)
        f.setVariableIndices(variableIndex)

    return elementaryExpressionIndex,allFreeBetas,freeBetaNames,allFixedBetas,fixedBetaNames,allRandomVariables,randomVariableNames,allDraws,drawNames



import os
import glob

def runAllScriptsInDir(dir):
    os.chdir(dir)
    print(f"*** Run tests in {dir} ***")
    for f in glob.glob("*.py"):
        if f != "__init__.py":
            print(f"Run test {f}")


listOfDirs = list()
for (dirpath, dirnames, filenames) in os.walk('.'):
    if '__pycache__' not in dirpath and dirpath != '.':
        listOfDirs += [dirpath]

for l in listOfDirs:
    runAllScriptsInDir(l)


#    print(f"dirpath={dirpath}")
#    print(f"dirnames={dirnames}")
#    for f in filenames:
#        if f.endswith(".py") and f != 'run_all.py':
#            listOfFiles += [os.path.join(dirpath, f)]

#for f in listOfFiles:
#    os.system(f"python {f}")



"""Defines a generic exception for Biogeme

:author: Michel Bierlaire

:date: Tue Mar 26 16:47:11 2019

"""
class biogemeError(Exception):
    """ Defines a generic exception for Biogeme. """ 
    pass



"""File 07problem.py

 This file is the same as 02oneLatentOrdered.py, where The starting
 values for the sigma have been change in order to illustrate a common
 issue with the estimation of such models.
 
 We set the starting value of a scale parameter (SIGMA_STAR_Envir02)
 to a small value: 0.01. The resulting likelihood is so close to zero
 that taking the log generates a numerical issue.

 Make sure to set large initial values for scale parameters.

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 15:40:43 2019

"""
import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
from biogeme.models import piecewise
import biogeme.loglikelihood as ll
from biogeme.expressions import Beta, DefineVariable, log, Elem, bioNormalCdf

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)

# Variables

# Piecewise linear definition of income
ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)
thresholds = [4,6,8,10]
ContIncome = piecewise(ScaledIncome,thresholds)
ContIncome_0_4000 = ContIncome[0]
ContIncome_4000_6000 = ContIncome[1]
ContIncome_6000_8000 = ContIncome[2]
ContIncome_8000_10000 = ContIncome[3]
ContIncome_10000_more = ContIncome[4]

# Definition of other variables
age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

# Coefficients
coef_intercept = Beta('coef_intercept',0.0,None,None,0 )
coef_age_65_more = Beta('coef_age_65_more',0.0,None,None,0 )
coef_haveGA = Beta('coef_haveGA',0.0,None,None,0 )
coef_ContIncome_0_4000 = \
 Beta('coef_ContIncome_0_4000',0.0,None,None,0 )
coef_ContIncome_4000_6000 = \
 Beta('coef_ContIncome_4000_6000',0.0,None,None,0 )
coef_ContIncome_6000_8000 = \
 Beta('coef_ContIncome_6000_8000',0.0,None,None,0 )
coef_ContIncome_8000_10000 = \
 Beta('coef_ContIncome_8000_10000',0.0,None,None,0 )
coef_ContIncome_10000_more = \
 Beta('coef_ContIncome_10000_more',0.0,None,None,0 )
coef_moreThanOneCar = \
 Beta('coef_moreThanOneCar',0.0,None,None,0 )
coef_moreThanOneBike = \
 Beta('coef_moreThanOneBike',0.0,None,None,0 )
coef_individualHouse = \
 Beta('coef_individualHouse',0.0,None,None,0 )
coef_male = Beta('coef_male',0.0,None,None,0 )
coef_haveChildren = Beta('coef_haveChildren',0.0,None,None,0 )
coef_highEducation = Beta('coef_highEducation',0.0,None,None,0 )

# Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

CARLOVERS = \
            coef_intercept +\
            coef_age_65_more * age_65_more +\
            coef_ContIncome_0_4000 * ContIncome_0_4000 +\
            coef_ContIncome_4000_6000 * ContIncome_4000_6000 +\
            coef_ContIncome_6000_8000 * ContIncome_6000_8000 +\
            coef_ContIncome_8000_10000 * ContIncome_8000_10000 +\
            coef_ContIncome_10000_more * ContIncome_10000_more +\
            coef_moreThanOneCar * moreThanOneCar +\
            coef_moreThanOneBike * moreThanOneBike +\
            coef_individualHouse * individualHouse +\
            coef_male * male +\
            coef_haveChildren * haveChildren +\
            coef_haveGA * haveGA +\
            coef_highEducation * highEducation


# Measurement equations

INTER_Envir01 = Beta('INTER_Envir01',0,None,None,1)
INTER_Envir02 = Beta('INTER_Envir02',0.0,None,None,0 )
INTER_Envir03 = Beta('INTER_Envir03',0.0,None,None,0 )
INTER_Mobil11 = Beta('INTER_Mobil11',0.0,None,None,0 )
INTER_Mobil14 = Beta('INTER_Mobil14',0.0,None,None,0 )
INTER_Mobil16 = Beta('INTER_Mobil16',0.0,None,None,0 )
INTER_Mobil17 = Beta('INTER_Mobil17',0.0,None,None,0 )

B_Envir01_F1 = Beta('B_Envir01_F1',-1,None,None,1)
B_Envir02_F1 = Beta('B_Envir02_F1',0.0,None,None,0 )
B_Envir03_F1 = Beta('B_Envir03_F1',0.0,None,None,0 )
B_Mobil11_F1 = Beta('B_Mobil11_F1',0.0,None,None,0 )
B_Mobil14_F1 = Beta('B_Mobil14_F1',0.0,None,None,0 )
B_Mobil16_F1 = Beta('B_Mobil16_F1',0.0,None,None,0 )
B_Mobil17_F1 = Beta('B_Mobil17_F1',0.0,None,None,0 )

MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01',1,None,None,1)

# We change the starting value of the following parameter compared to
# 02oneLatentOrdered.py
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02',0.01,None,None,0 )
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03',1,None,None,0 )
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11',1,None,None,0 )
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14',1,None,None,0 )
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16',1,None,None,0 )
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17',1,None,None,0 )

delta_1 = Beta('delta_1',0.1,0,10,0 )
delta_2 = Beta('delta_2',0.2,0,10,0 )
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1 
tau_3 = delta_1
tau_4 = delta_1 + delta_2

Envir01_tau_1 = (tau_1-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_2 = (tau_2-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_3 = (tau_3-MODEL_Envir01) / SIGMA_STAR_Envir01
Envir01_tau_4 = (tau_4-MODEL_Envir01) / SIGMA_STAR_Envir01
IndEnvir01 = {
    1: bioNormalCdf(Envir01_tau_1),
    2: bioNormalCdf(Envir01_tau_2)-bioNormalCdf(Envir01_tau_1),
    3: bioNormalCdf(Envir01_tau_3)-bioNormalCdf(Envir01_tau_2),
    4: bioNormalCdf(Envir01_tau_4)-bioNormalCdf(Envir01_tau_3),
    5: 1-bioNormalCdf(Envir01_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir01 = Elem(IndEnvir01, Envir01)


Envir02_tau_1 = (tau_1-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_2 = (tau_2-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_3 = (tau_3-MODEL_Envir02) / SIGMA_STAR_Envir02
Envir02_tau_4 = (tau_4-MODEL_Envir02) / SIGMA_STAR_Envir02
IndEnvir02 = {
    1: bioNormalCdf(Envir02_tau_1),
    2: bioNormalCdf(Envir02_tau_2)-bioNormalCdf(Envir02_tau_1),
    3: bioNormalCdf(Envir02_tau_3)-bioNormalCdf(Envir02_tau_2),
    4: bioNormalCdf(Envir02_tau_4)-bioNormalCdf(Envir02_tau_3),
    5: 1-bioNormalCdf(Envir02_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir02 = Elem(IndEnvir02, Envir02)

Envir03_tau_1 = (tau_1-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_2 = (tau_2-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_3 = (tau_3-MODEL_Envir03) / SIGMA_STAR_Envir03
Envir03_tau_4 = (tau_4-MODEL_Envir03) / SIGMA_STAR_Envir03
IndEnvir03 = {
    1: bioNormalCdf(Envir03_tau_1),
    2: bioNormalCdf(Envir03_tau_2)-bioNormalCdf(Envir03_tau_1),
    3: bioNormalCdf(Envir03_tau_3)-bioNormalCdf(Envir03_tau_2),
    4: bioNormalCdf(Envir03_tau_4)-bioNormalCdf(Envir03_tau_3),
    5: 1-bioNormalCdf(Envir03_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Envir03 = Elem(IndEnvir03, Envir03)

Mobil11_tau_1 = (tau_1-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_2 = (tau_2-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_3 = (tau_3-MODEL_Mobil11) / SIGMA_STAR_Mobil11
Mobil11_tau_4 = (tau_4-MODEL_Mobil11) / SIGMA_STAR_Mobil11
IndMobil11 = {
    1: bioNormalCdf(Mobil11_tau_1),
    2: bioNormalCdf(Mobil11_tau_2)-bioNormalCdf(Mobil11_tau_1),
    3: bioNormalCdf(Mobil11_tau_3)-bioNormalCdf(Mobil11_tau_2),
    4: bioNormalCdf(Mobil11_tau_4)-bioNormalCdf(Mobil11_tau_3),
    5: 1-bioNormalCdf(Mobil11_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil11 = Elem(IndMobil11, Mobil11)

Mobil14_tau_1 = (tau_1-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_2 = (tau_2-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_3 = (tau_3-MODEL_Mobil14) / SIGMA_STAR_Mobil14
Mobil14_tau_4 = (tau_4-MODEL_Mobil14) / SIGMA_STAR_Mobil14
IndMobil14 = {
    1: bioNormalCdf(Mobil14_tau_1),
    2: bioNormalCdf(Mobil14_tau_2)-bioNormalCdf(Mobil14_tau_1),
    3: bioNormalCdf(Mobil14_tau_3)-bioNormalCdf(Mobil14_tau_2),
    4: bioNormalCdf(Mobil14_tau_4)-bioNormalCdf(Mobil14_tau_3),
    5: 1-bioNormalCdf(Mobil14_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil14 = Elem(IndMobil14, Mobil14)

Mobil16_tau_1 = (tau_1-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_2 = (tau_2-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_3 = (tau_3-MODEL_Mobil16) / SIGMA_STAR_Mobil16
Mobil16_tau_4 = (tau_4-MODEL_Mobil16) / SIGMA_STAR_Mobil16
IndMobil16 = {
    1: bioNormalCdf(Mobil16_tau_1),
    2: bioNormalCdf(Mobil16_tau_2)-bioNormalCdf(Mobil16_tau_1),
    3: bioNormalCdf(Mobil16_tau_3)-bioNormalCdf(Mobil16_tau_2),
    4: bioNormalCdf(Mobil16_tau_4)-bioNormalCdf(Mobil16_tau_3),
    5: 1-bioNormalCdf(Mobil16_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil16 = Elem(IndMobil16, Mobil16)

Mobil17_tau_1 = (tau_1-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_2 = (tau_2-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_3 = (tau_3-MODEL_Mobil17) / SIGMA_STAR_Mobil17
Mobil17_tau_4 = (tau_4-MODEL_Mobil17) / SIGMA_STAR_Mobil17
IndMobil17 = {
    1: bioNormalCdf(Mobil17_tau_1),
    2: bioNormalCdf(Mobil17_tau_2)-bioNormalCdf(Mobil17_tau_1),
    3: bioNormalCdf(Mobil17_tau_3)-bioNormalCdf(Mobil17_tau_2),
    4: bioNormalCdf(Mobil17_tau_4)-bioNormalCdf(Mobil17_tau_3),
    5: 1-bioNormalCdf(Mobil17_tau_4),
    6: 1.0,
    -1: 1.0,
    -2: 1.0
}

P_Mobil17 = Elem(IndMobil17, Mobil17)


loglike = log(P_Envir01) + \
          log(P_Envir02) + \
          log(P_Envir03) + \
          log(P_Mobil11) + \
          log(P_Mobil14) + \
          log(P_Mobil16) + \
          log(P_Mobil17)

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,loglike)
biogeme.modelName = "07problem"

# Estimate the parameters
results = biogeme.estimate()

print(f"Estimated betas: {len(results.data.betaValues)}")
print(f"final log likelihood: {results.data.logLike:.3f}")
print(f"Output file: {results.data.htmlFileName}")
results.writeLaTeX()
print(f"LaTeX file: {results.data.latexFileName}")






import os
import sys
import platform
if platform.system() == 'Darwin':
    os.environ["CC"] = 'clang'
    os.environ["CXX"] = 'clang++'
else:
    os.environ["CC"] = 'gcc'
    os.environ["CXX"] = 'g++'

#from distutils.core import setup
#from distutils.extension import Extension
import numpy
import setuptools
from setuptools import setup, Extension, Command
from setuptools.dist import Distribution
import setuptools.command.test
from setuptools import (find_packages, setup)




# monkey-patch for parallel compilation
def parallelCCompile(self, sources, output_dir=None, macros=None, include_dirs=None, debug=0, extra_preargs=None, extra_postargs=None, depends=None):
    # those lines are copied from distutils.ccompiler.CCompiler directly
    macros, objects, extra_postargs, pp_opts, build = self._setup_compile(output_dir, macros, include_dirs, sources, depends, extra_postargs)
    cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
    # parallel code
    N=8 # number of parallel compilations
    import multiprocessing.pool
    def _single_compile(obj):
        try: src, ext = build[obj]
        except KeyError: return
        self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
    # convert to list, imap is evaluated on-demand
    list(multiprocessing.pool.ThreadPool(N).imap(_single_compile,objects))
    return objects

import distutils.ccompiler
distutils.ccompiler.CCompiler.compile=parallelCCompile

try:
    from Cython.Build import cythonize
    from Cython.Distutils import build_ext
except ImportError:
    USE_CYTHON= False
    print("Not using Cython")
else:
    USE_CYTHON = True
    print("Using Cython")

ext = '.pyx' if USE_CYTHON else '.cpp'

cmdclass = {}

source = ["src/biogeme.cc",
          "src/bioNormalCdf.cc",
          "src/bioFormula.cc",
          "src/bioThreadMemory.cc",
          "src/bioString.cc",
          "src/bioExprNormalCdf.cc",
          "src/bioExprIntegrate.cc",
          "src/bioExprGaussHermite.cc",
          "src/bioExprRandomVariable.cc",
          "src/bioExprMontecarlo.cc",
          "src/bioExprPanelTrajectory.cc",
          "src/bioExprDraws.cc",
          "src/bioExprDerive.cc",
          "src/bioExprMin.cc",
          "src/bioExprMax.cc",
          "src/bioExprAnd.cc",
          "src/bioExprOr.cc",
          "src/bioExprEqual.cc",
          "src/bioExprNotEqual.cc",
          "src/bioExprLessOrEqual.cc",
          "src/bioExprLess.cc",
          "src/bioExprGreaterOrEqual.cc",
          "src/bioExprGreater.cc",
          "src/bioExprElem.cc",
          "src/bioExprMultSum.cc",
          "src/bioExprLiteral.cc",
          "src/bioExprFreeParameter.cc",
          "src/bioExprFixedParameter.cc",
          "src/bioExprVariable.cc",
          "src/bioExprPlus.cc",
          "src/bioExprMinus.cc",
          "src/bioExprTimes.cc",
          "src/bioExprDivide.cc",
          "src/bioExprPower.cc",
          "src/bioExprUnaryMinus.cc",
          "src/bioExprExp.cc",
          "src/bioExprLog.cc",
          "src/bioExprNumeric.cc",
          "src/bioExprLogLogit.cc",
          "src/bioExprLogLogitFullChoiceSet.cc",
          "src/bioExprLinearUtility.cc",
          "src/bioExpression.cc",
          "src/bioExceptions.cc",
          "src/bioDerivatives.cc",
          "src/bioGaussHermite.cc",
          "src/bioGhFunction.cc",
          "src/mycfsqp.cc",
          "src/myqld.cc",
          "src/bioCfsqp.cc"
]


extra_compile_args = ["-std=c++11","-Wall"]
extra_link_args = []

if platform.system() == 'Darwin':
	extra_compile_args.append("-stdlib=libc++")
	extra_link_args.append("-lc++")

if sys.platform == "win32":
	# mismatch between library names
	extra_compile_args.append("-D_hypot=hypot")
	#as one cannot assume that libstdc++, libgcc and pthreads exists on windows, static link them so they are included in the compiled python extension.
	extra_link_args.extend(['-mwindows','-mms-bitfields','-static-libstdc++', '-static-libgcc', '-Wl,-Bstatic', '-lstdc++', '-lpthread'])

extensions = [Extension("biogeme.cbiogeme", ["src/cbiogeme"+ext]+source,include_dirs=[numpy.get_include()],extra_compile_args=extra_compile_args, language='c++11', extra_link_args = extra_link_args)]

#extensions = [Extension("biogeme.cbiogeme", ["src/cbiogeme"+ext]+source,include_dirs=[numpy.get_include()],extra_compile_args=extra_compile_args,language='c++11', extra_link_args = ["-lprofiler"])]


#extra_compile_args=["-O0"],
#extra_link_args=["-fsanitize=address","-O1","-fno-omit-frame-pointer","-g"]

if USE_CYTHON:
    extensions = cythonize(extensions,language="c++",include_path=[numpy.get_include()])
    cmdclass.update({ 'build_ext': build_ext})

exec(open("biogeme/version.py").read())
# now we have a `__version__` variable
# later on we use: __version__    

with open("README.md", "r") as fh:
    long_description = fh.read()

class BinaryDistribution(Distribution):
   """Distribution which always forces a binary package with platform name"""
   def has_ext_modules(foo):
      return True

setup(name = "biogeme",
      version=__version__,
      description="Estimation and application of discrete choice models",
      url="http://biogeme.epfl.ch",
      author="Michel Bierlaire",
      keywords="discrete choice maximum likelihood estimation",
      author_email="michel.bierlaire@epfl.ch",
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=['numpy','cython','unidecode','scipy','pandas'],
      packages=setuptools.find_packages(),
      include_package_data=True,
      package_data={"biogeme": ["_biogeme.pyd"]},
      package_dir = {"biogeme.cbiogeme":"src"},
      cmdclass = cmdclass,
      classifiers=[
          "Programming Language :: Python :: 3",
          "Operating System :: OS Independent",
      ],
      ext_modules = extensions
)






"""File: 01simpleIntegral.py
 Author: Michel Bierlaire, EPFL
 Date: Wed Dec 11 16:20:24 2019

Calculation of a simple integral using Monte-Carlo integration.

"""

import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.draws as draws
from biogeme.expressions import exp, bioDraws, MonteCarlo

# We create a fake database with one entry, as it is required
# to store the draws
pandas = pd.DataFrame()
pandas['FakeColumn'] = [1.0]
database = db.Database('fakeDatabase',pandas)

integrand = exp(bioDraws('U','UNIFORM'))
simulatedI = MonteCarlo(integrand)

trueI = exp(1.0) - 1.0 

R = 2000

sampleVariance = \
  MonteCarlo(integrand*integrand) - simulatedI * simulatedI
stderr = (sampleVariance / R)**0.5
error = simulatedI - trueI

simulate = {'Analytical Integral': trueI,
            'Simulated Integral': simulatedI,
            'Sample variance   ': sampleVariance,
            'Std Error         ': stderr,
            'Error             ': error}

biogeme = bio.BIOGEME(database,simulate,numberOfDraws=R)
biogeme.modelName = f'01simpleIntegral_{R}'
results = biogeme.simulate()
print(f'Number of draws: {R}')
for c in results.columns:
    print(f'{c}: {results.loc[0,c]}')

# With 10 times more draws
biogeme2 = bio.BIOGEME(database,simulate,numberOfDraws=10*R)
biogeme2.modelName = '01simpleIntegral_{10*R}'
results2 = biogeme2.simulate()
print(f'Number of draws: {10*R}')
for c in results.columns:
    print(f'{c}: {results2.loc[0,c]}')






"""File 12panel.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 18:55:38 2019

 Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable, bioDraws, PanelLikelihoodTrajectory, MonteCarlo, log

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# They are organized as panel data. The variable ID identifies each individual.
database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)
B_TIME_S = Beta('B_TIME_S',0,None,None,0)

# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND','NORMAL')

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# Definition of the utility functions
V1 = ASC_TRAIN + \
     B_TIME_RND * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME_RND * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME_RND * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# Conditional to B_TIME_RND, the likelihood of one observation is
# given by the logit model (called the kernel)
obsprob = models.logit(V,av,CHOICE)

# Conditional to B_TIME_RND, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# We integrate over B_TIME_RND using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
#logger.setGeneral()
#logger.setDetailed()
logger.setDebug()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob,numberOfDraws=50)
biogeme.modelName = "12panel"

# Estimate the parameters. 
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)






"""File 02nestedPlot.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 10:15:18 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We simulate pricing scenarios and their impact on the revenues.
"""
import sys
import numpy as np
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable

# Library for plotting
import matplotlib.pyplot as plt

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)

# Define new variables
TimePT_scaled = DefineVariable('TimePT_scaled', TimePT / 200 ,database)
TimeCar_scaled = DefineVariable('TimeCar_scaled', TimeCar / 200 ,database)
CostCarCHF_scaled = DefineVariable('CostCarCHF_scaled',
                                   CostCarCHF / 10 ,database)
distance_km_scaled = DefineVariable('distance_km_scaled',
                                    distance_km / 5 ,database)
male = DefineVariable('male',Gender == 1,database)
female = DefineVariable('female',Gender == 2,database)
unreportedGender = DefineVariable('unreportedGender',Gender == -1,database)
fulltime = DefineVariable('fulltime',OccupStat == 1,database)
notfulltime = DefineVariable('notfulltime',OccupStat != 1,database)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
normalizedWeight = Weight * 1906 / 0.814484

 
def scenario(scale):
    """Simulate a scenarios modifying the price of public transportation

    :param scale: price multiplier.
    :type scale: float
    :return: simulated revenues
    :rtype: float
    """
    # This is the only variable that depends on scale
    MarginalCostScenario = MarginalCostPT * scale
    MarginalCostPT_scaled = MarginalCostScenario / 10
    # The rest of the model is the same for all scenarios
    ASC_CAR = Beta('ASC_CAR',0,None,None,0)
    ASC_PT = Beta('ASC_PT',0,None,None,1)
    ASC_SM = Beta('ASC_SM',0,None,None,0)
    BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME',0,None,None,0)
    BETA_TIME_OTHER = Beta('BETA_TIME_OTHER',0,None,None,0)
    BETA_DIST_MALE = Beta('BETA_DIST_MALE',0,None,None,0)
    BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE',0,None,None,0)
    BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED',0,None,None,0)
    BETA_COST = Beta('BETA_COST',0,None,None,0)
    # Utility functions
    V_PT = ASC_PT + BETA_TIME_FULLTIME * TimePT_scaled * fulltime + \
           BETA_TIME_OTHER * TimePT_scaled * notfulltime + \
           BETA_COST * MarginalCostPT_scaled
    V_CAR = ASC_CAR + \
            BETA_TIME_FULLTIME * TimeCar_scaled * fulltime + \
            BETA_TIME_OTHER * TimeCar_scaled * notfulltime + \
            BETA_COST * CostCarCHF_scaled
    V_SM = ASC_SM + \
           BETA_DIST_MALE * distance_km_scaled * male + \
           BETA_DIST_FEMALE * distance_km_scaled * female + \
           BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender
    V = {0: V_PT,
         1: V_CAR,
         2: V_SM}
    av = {0: 1,
          1: 1,
          2: 1}
    MU_NOCAR = Beta('MU_NOCAR',1.0,1.0,None,0)
    CAR_NEST = 1.0 , [ 1]
    NO_CAR_NEST = MU_NOCAR , [ 0, 2]
    nests = CAR_NEST, NO_CAR_NEST
    prob_pt = models.nested(V,av,nests,0)
    simulate = {'weight': normalizedWeight,
                'Revenue public transportation':
                   prob_pt * MarginalCostScenario}

    biogeme  = bio.BIOGEME(database,simulate)

    # Retrieve the values of the parameters.  First, extract the names
    # of parameters needed for the simulation
    betas = biogeme.freeBetaNames

    # Read the estimation results from the file
    results = res.bioResults(pickleFile='01nestedEstimation.pickle')

    # Extract the values that are necessary
    betaValues = results.getBetaValues(betas)
    
    simulatedValues = biogeme.simulate(betaValues)

    # We calculate the sum for all individuals of the generated revenues.
    revenues_pt = (simulatedValues['Revenue public transportation'] * simulatedValues['weight']).sum()
    return revenues_pt

# We now plot the relationship between the cost and the revenues
scales = np.arange(0.0,5.0,0.1)
revenues = [scenario(s) for s in scales]
plt.plot(scales,revenues)
plt.xlabel("Modification of the price of public transportation (%)")
plt.ylabel("Revenues")
plt.show()



import unittest
import random as rnd
import biogeme.biogeme as bio
import biogeme.database as db
import biogeme.models as models
import biogeme.optimization as opt
import pandas as pd
import numpy as np
import biogeme.exceptions as excep
from biogeme.expressions import *

class rosenbrock(opt.functionToMinimize):
    def __init__(self):
        self.x = None
        
    def setVariables(self,x):
        self.x = x

    def f(self, batch = None):
        if batch is not None:
            raise excep.biogemeError('This function is not data driven.')
        n = len(self.x)
        f = sum(100.0 * (self.x[i+1]-self.x[i]**2)**2 + (1.0-self.x[i])**2 for i in range(n-1))
        return f

    def g(self):
        n = len(self.x)
        g = np.zeros(n)
        for i in range(n-1):
            g[i] = g[i] - 400 * self.x[i] * (self.x[i+1]-self.x[i]**2) - 2 * (1-self.x[i])
            g[i+1] = g[i+1] + 200 * (self.x[i+1]-self.x[i]**2)

        return g
    def h(self):
        n = len(self.x)
        H = np.zeros((n,n))
        for i in range(n-1):
            H[[i],[i]] = H[[i],[i]] - 400 * self.x[i+1] + 1200 * self.x[i]**2 + 2
            H[[i+1],[i]] = H[[i+1],[i]] - 400 * self.x[i]
            H[[i],[i+1]] = H[[i],[i+1]] - 400 * self.x[i]
            H[[i+1],[i+1]] = H[[i+1],[i+1]] + 200
        return H 

    def f_g(self, batch = None):
        if batch is not None:
            raise excep.biogemeError('This function is not data driven.')
        return self.f(), self.g()

    def f_g_h(self, batch= None):
        if batch is not None:
            raise excep.biogemeError('This function is not data driven.')
        return self.f(), self.g(), self.h()
    
    def f_g_bhhh(self, batch = None):
        raise excep.biogemeError('This function is not data driven.')
        


class testOptimization(unittest.TestCase):
    def setUp(self):
        np.random.seed(90267)
        rnd.seed(90267)
        df = pd.DataFrame({'Person':[1,1,1,2,2],
                   'Exclude':[0,0,1,0,1],
                   'Variable1':[1,2,3,4,5],
                   'Variable2':[10,20,30,40,50],
                   'Choice':[1,2,3,1,2],
                   'Av1':[0,1,1,1,1],
                   'Av2':[1,1,1,1,1],
                   'Av3':[0,1,1,1,1]})
        myData = db.Database('test',df)
        
        Choice=Variable('Choice')
        Variable1=Variable('Variable1')
        Variable2=Variable('Variable2')
        beta1 = Beta('beta1',0,None,None,0)
        beta2 = Beta('beta2',0,None,None,0)
        V1 = beta1 * Variable1
        V2 = beta2 * Variable2
        V3 = 0
        V ={1:V1,2:V2,3:V3}

        likelihood = models.loglogit(V,av=None,i=Choice)
        self.myBiogeme = bio.BIOGEME(myData,likelihood)
        self.myBiogeme.modelName = 'simpleExample'
        self.theFunction = rosenbrock()


    def testSchnabelEskow(self):
        A = np.array([[0.3571,-0.1030,0.0274,-0.0459],[-0.1030,0.2525,0.0736,-0.3845],[0.0274,0.0736,0.2340,-0.2878],[-0.0459,-0.3845,-0.2878,0.5549]])
        L,E,P = opt.schnabelEskow(A)
        self.assertAlmostEqual(L[0,0],0.7449161,5)
        diff = P @ L @ L.T @ P.T - E - A
        for i in list(diff.flatten()):
            self.assertAlmostEqual(i,0.0,5)

    def testSchnabelEskow2(self):
        A = np.array([[1890.3,-1705.6,-315.8,3000.3],[-1705.6,1538.3,284.9,-2706.6],[-315.8,284.9,52.5,-501.2],[3000.3,-2706.6,-501.2,4760.8]])        
        L,E,P = opt.schnabelEskow(A)
        self.assertAlmostEqual(L[0,0],6.89985507e+01,5)
        diff = P @ L @ L.T @ P.T - E - A
        for i in list(diff.flatten()):
            self.assertAlmostEqual(i,0.0,5)

    def testLineSearch(self):
        x = np.array([-1.5,1.5])
        self.theFunction.setVariables(x)
        f,g = self.theFunction.f_g()
        alpha,nfev = opt.lineSearch(self.theFunction,x,f,g,-g)
        self.assertAlmostEqual(alpha,0.0009765625)

    def testNewton(self):
        x0 = np.array([-1.5,1.5])
        xstar,nit,nfev,msg = opt.newtonLineSearch(self.theFunction,x0)
        for i in list(xstar):
            self.assertAlmostEqual(i,1,4)

    def testNewtonTrustRegion(self):
        x0 = np.array([-1.5,1.5])
        xstar,nit,nfev,msg = opt.newtonTrustRegion(self.theFunction,x0)
        for i in list(xstar):
            self.assertAlmostEqual(i,1,4)
        
    def testBioScipy(self):
        results = self.myBiogeme.estimate()
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'],0.144546,3)
        self.assertAlmostEqual(beta['beta2'],0.023502,3)

    def testBioNewtonLineSearch(self):
        results = self.myBiogeme.estimate(algorithm=opt.newtonLineSearchForBiogeme)
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'],0.144546,3)
        self.assertAlmostEqual(beta['beta2'],0.023502,3)

    def testBioCfsqp(self):
        results = self.myBiogeme.estimate(algorithm=None)
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'],0.144546,3)
        self.assertAlmostEqual(beta['beta2'],0.023502,3)

    def testBioNewtonTrustRegion(self):
        results = self.myBiogeme.estimate(algorithm=opt.newtonTrustRegionForBiogeme)
        beta = results.getBetaValues()
        self.assertAlmostEqual(beta['beta1'],0.144546,3)
        self.assertAlmostEqual(beta['beta2'],0.023502,3)
        
if __name__ == '__main__':
    unittest.main()



import biogeme.results as res


structResults = res.bioResults(pickleFile='05latentChoiceFull.pickle')
#structResults = res.bioResults(pickleFile='04latentChoiceSeq.pickle')
#structResults = res.bioResults(pickleFile='02oneLatentOrdered.pickle')
structBetas = structResults.getBetaValues()

for k,v in structBetas.items():
    print(f"{k} = Beta('{k}',{v},None,None,0)")

#for k,v in structBetas.items():
#    print(f"{k} = {v}")
    



import pandas as pd
import sys
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme.expressions import Beta, DefineVariable, Elem, log
import unittest


pandas = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",pandas)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)


B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)

tau1	 = Beta('tau1',-1,None,0,0)
delta2	 = Beta('delta2',2,0,None,0)

tau2 = tau1 + delta2


TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)

#  Utility

U = B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED 


ChoiceProba = {
    1: 1-dist.logisticcdf(U-tau1),
    2: dist.logisticcdf(U-tau1)- dist.logisticcdf(U-tau2),
    3: dist.logisticcdf(U-tau2) }

logprob = log(Elem(ChoiceProba,CHOICE))

class test_18(unittest.TestCase):
    def testEstimation(self):
        biogeme  = bio.BIOGEME(database,logprob)
        biogeme.modelName = "test_18"
        results = biogeme.estimate(algorithm=None)
        self.assertAlmostEqual(results.data.logLike,-5789.309,2)
    
if __name__ == '__main__':
    unittest.main()


